"""
Recursively Walk Into Directories and Archives
==============================================

This module primarily provides the function :func:`unzipwalk`, which recursively walks into
directories and compressed files and returns all files, directories, etc. found, together with
binary file handles (file objects) for reading the files. Currently supported are ``zip``, ``gz``,
``bz2``, ``xz``, and the various ``tar`` compressed formats ``tar.gz``/``tgz``, ``tar.xz``/``txz``,
and ``tar.bz2``/``tbz``/``tbz2``, plus ``7z`` files if the Python package :mod:`py7zr` is installed
(to get the latter, you can install this package with ``pip install unzipwalk[7z]``). File types
are detected based on the aforementioned extensions.

    >>> from unzipwalk import unzipwalk
    >>> results = []
    >>> for result in unzipwalk('.'):
    ...     names = tuple( name.as_posix() for name in result.names )
    ...     if result.hnd:  # result is a file opened for reading (binary)
    ...         # could use result.hnd.read() here, or for line-by-line:
    ...         for line in result.hnd:
    ...             pass  # do something interesting with the data here
    ...     results.append(names + (result.typ.name,))
    >>> print(sorted(results))# doctest: +NORMALIZE_WHITESPACE
    [('bar.zip', 'ARCHIVE'),
     ('bar.zip', 'bar.txt', 'FILE'),
     ('bar.zip', 'test.tar.gz', 'ARCHIVE'),
     ('bar.zip', 'test.tar.gz', 'hello.csv', 'FILE'),
     ('bar.zip', 'test.tar.gz', 'test', 'DIR'),
     ('bar.zip', 'test.tar.gz', 'test/cool.txt.gz', 'ARCHIVE'),
     ('bar.zip', 'test.tar.gz', 'test/cool.txt.gz', 'test/cool.txt', 'FILE'),
     ('foo.txt', 'FILE')]

**Note** that :func:`unzipwalk` automatically closes files as it goes from file to file.
This means that you must use the handles as soon as you get them from the generator -
something as seemingly simple as ``sorted(unzipwalk('.'))`` would cause the code above to fail,
because all files will have been opened and closed during the call to :func:`sorted`
and the handles to read the data would no longer be available in the body of the loop.
This is why the above example first processes all the files before sorting the results.
You can also use :func:`recursive_open` to open the files later, though using that function
is less efficient that :func:`unzipwalk` if you are opening multiple files inside of Zip
or tar archives.

The yielded file handles can be wrapped in :class:`io.TextIOWrapper` to read them as text files.
For example, to read all CSV files in the current directory and below, including within compressed files:

    >>> from unzipwalk import unzipwalk, FileType
    >>> from io import TextIOWrapper
    >>> import csv
    >>> for result in unzipwalk('.'):
    ...     if result.typ==FileType.FILE and result.names[-1].suffix.lower()=='.csv':
    ...         print([ name.as_posix() for name in result.names ])
    ...         with TextIOWrapper(result.hnd, encoding='UTF-8', newline='') as handle:
    ...             csv_rd = csv.reader(handle, strict=True)
    ...             for row in csv_rd:
    ...                 print(repr(row))
    ['bar.zip', 'test.tar.gz', 'hello.csv']
    ['Id', 'Name', 'Address']
    ['42', 'Hello', 'World']

.. note::
    The original names of files compressed with gzip, bzip2, and lzma are derived by
    simply removing the respective ``.gz``, ``.bz2``, or ``.xz`` extensions.

    Using the original filename from the gzip file's header is currently not possible due to
    `limitations in the underlying library <https://github.com/python/cpython/issues/71638>`_.

.. seealso::
    - `zipfile Issues <https://github.com/orgs/python/projects/7>`_
    - `tarfile Issues <https://github.com/orgs/python/projects/11>`_
    - `Compression issues <https://github.com/orgs/python/projects/20>`_ (gzip, bzip2, lzma)
    - `py7zr Issues <https://github.com/miurahr/py7zr/issues>`_

API
---

.. autofunction:: unzipwalk.unzipwalk

.. autoclass:: unzipwalk.UnzipWalkResult
    :members:

.. autoclass:: unzipwalk.FileType
    :members:

.. autofunction:: unzipwalk.recursive_open

.. autoclass:: unzipwalk.ReadOnlyBinary
    :members:
    :undoc-members:

Command-Line Interface
----------------------

.. unzipwalk_clidoc::

The available checksum algorithms may vary depending on your system and Python version.
Run the command with ``--help`` to see the list of algorithms available on your system.

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import io
import re
import stat
from bz2 import BZ2File
from gzip import GzipFile
from lzma import LZMAFile
from tarfile import TarFile
from zipfile import ZipFile
from contextlib import contextmanager
from pathlib import PurePosixPath, Path
from typing import Optional, cast, IO, Union
from collections.abc import Generator, Sequence
from igbpyutils.file import AnyPaths, to_Paths, Filename
from .defs import FileType, UnzipWalkResult, ReadOnlyBinary, FilterType, FileProcessorArgs, ProcessCallContext, RecursiveOpenArgs

# spell-checker: ignore autoclass autofunction clidoc seealso undoc

__all__ = ['FileType', 'UnzipWalkResult', 'ReadOnlyBinary', 'FilterType', 'recursive_open', 'unzipwalk']

try:  # cover-req-lt3.14
    from .wrap7z import Wrap7Z
    W7Z :Optional[Wrap7Z] = Wrap7Z()  # pylint: disable=invalid-name,useless-suppression
except (ImportError, OSError):  # cover-req-ge3.14
    W7Z = None  # pylint: disable=invalid-name,useless-suppression  # pyright: ignore [reportConstantRedefinition]

_TARFILE_RE = re.compile(r'\.(?:tar(?:\.gz|\.bz2|\.xz)?|tgz|txz|tbz2?)\Z', re.I)

@contextmanager
def _inner_recur_open(a: RecursiveOpenArgs) -> Generator[IO[bytes], None, None]:
    try:
        bl = a.fns[0].name.lower()
        assert a.fns, a.fns
        if len(a.fns)==1:
            yield a.fh
        # the following code is very similar to _proc_file, please see those code comments for details
        elif _TARFILE_RE.search(bl):
            with TarFile.open(fileobj=a.fh) as tf:
                ef = tf.extractfile(str(a.fns[1]))
                if not ef:  # e.g. directory
                    raise FileNotFoundError(f"not a file? {a.fns[0:2]}")  # [0]=the current file, [1]=the file we're trying to open
                #TODO Later: I'm not sure why the following two branch coverage exceptions are necessary on 3.14?
                with ef as fh2:
                    with _inner_recur_open(RecursiveOpenArgs(fns=a.fns[1:], fh=fh2)) as inner:  # pragma: no branch
                        yield inner
        elif bl.endswith('.zip'):
            with ZipFile(a.fh) as zf:
                with zf.open(str(a.fns[1])) as fh2:
                    with _inner_recur_open(RecursiveOpenArgs(fns=a.fns[1:], fh=fh2)) as inner:  # pragma: no branch
                        yield inner
        elif bl.endswith('.7z'):
            if W7Z:  # cover-req-lt3.14
                yield from W7Z.recursive_open(a, _inner_recur_open)
            else:  # cover-req-ge3.14
                raise ImportError("The py7zr package must be installed to open 7z files.")
        elif bl.endswith('.bz2'):
            if a.fns[1] != a.fns[0].with_suffix(''):
                raise FileNotFoundError(f"invalid bz2 filename {a.fns[0]} => {a.fns[1]}")
            with BZ2File(a.fh, mode='rb') as fh2:
                with _inner_recur_open(RecursiveOpenArgs(fns=a.fns[1:], fh=fh2)) as inner:
                    yield inner
        elif bl.endswith('.xz'):
            if a.fns[1] != a.fns[0].with_suffix(''):
                raise FileNotFoundError(f"invalid xz filename {a.fns[0]} => {a.fns[1]}")
            with LZMAFile(a.fh, mode='rb') as fh2:
                with _inner_recur_open(RecursiveOpenArgs(fns=a.fns[1:], fh=fh2)) as inner:
                    yield inner
        elif bl.endswith('.gz'):
            if a.fns[1] != a.fns[0].with_suffix(''):
                raise FileNotFoundError(f"invalid gzip filename {a.fns[0]} => {a.fns[1]}")
            with GzipFile(fileobj=a.fh, mode='rb') as fh2:
                with _inner_recur_open(RecursiveOpenArgs(fns=a.fns[1:], fh=cast(IO[bytes], fh2))) as inner:
                    yield inner
        else:
            assert False, 'should be unreachable: not all file types covered?'  # pragma: no cover
    except GeneratorExit:  # https://pylint.readthedocs.io/en/latest/user_guide/messages/warning/contextmanager-generator-missing-cleanup.html
        pass  # pragma: no cover

@contextmanager
def recursive_open(fns :Sequence[Filename], encoding=None, errors=None, newline=None) \
        -> Generator[Union[ReadOnlyBinary, io.TextIOWrapper], None, None]:
    # note Sphinx's "WARNING: py:class reference target not found: _io.TextIOWrapper" can be ignored
    """This context manager allows opening files nested inside archives directly.

    :func:`unzipwalk` automatically closes files as it iterates through directories and archives;
    this function exists to allow you to open the returned files after the iteration.
    However, this function will be less efficient that :func:`unzipwalk` if you're opening
    multiple files inside of Zip or tar archives.

    .. note: If *any* of ``encoding``, ``errors``, or ``newline`` is specified, the returned
        file is wrapped in :class:`io.TextIOWrapper`!

    .. note: If the last file in the list of files is an archive file, then it won't be decompressed,
        instead you'll be able to read the archive's raw compressed data from the handle.

    In this example, we open a gzip-compressed file, stored inside a tgz archive, which
    in turn is stored in a Zip file:

    >>> from unzipwalk import recursive_open
    >>> with recursive_open(('bar.zip', 'test.tar.gz', 'test/cool.txt.gz', 'test/cool.txt'), encoding='UTF-8') as fh:
    ...     print(fh.read())# doctest: +NORMALIZE_WHITESPACE
    Hi, I'm a compressed file!

    :raises ImportError: If you try to open a 7z file but :mod:`py7zr` is not installed.
    :raises Exception: See description in :func:`unzipwalk`.
    """
    if not fns:
        raise ValueError('no filenames given')
    with open(fns[0], 'rb') as fh:
        with _inner_recur_open(RecursiveOpenArgs(fns=(Path(fns[0]),) + tuple( PurePosixPath(f) for f in fns[1:] ), fh=fh)) as inner:
            assert inner.readable(), inner
            if encoding is not None or errors is not None or newline is not None:
                yield io.TextIOWrapper(inner, encoding=encoding, errors=errors, newline=newline)
            else:
                yield inner

def _proc_file(a :FileProcessorArgs) -> Generator[UnzipWalkResult, None, None]:  # pylint: disable=too-many-statements,too-many-branches
    bl = a.fns[-1].name.lower()
    if _TARFILE_RE.search(bl):
        try:
            with TarFile.open(fileobj=a.fh, errorlevel=2) as tf:
                for ti in tf.getmembers():
                    new_names = (*a.fns, PurePosixPath(ti.name))
                    if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
                        yield UnzipWalkResult(names=new_names, typ=FileType.SKIP)
                    # for ti.type see e.g.: https://github.com/python/cpython/blob/v3.12.3/Lib/tarfile.py#L88
                    elif ti.issym():
                        yield UnzipWalkResult(names=new_names, typ=FileType.SYMLINK)
                    elif ti.isdir():
                        yield UnzipWalkResult(names=new_names, typ=FileType.DIR)
                    elif ti.isfile():
                        try:
                            # Note apparently this can burn a lot of memory on <3.13: https://github.com/python/cpython/issues/102120
                            ef = tf.extractfile(ti)  # always binary
                            assert ef is not None, ti  # make type checker happy; we know this is true because we checked it's a file
                            with ef as fh2:
                                yield from _proc_file(FileProcessorArgs(fns=new_names, fh=fh2, size=ti.size, ctx=a.ctx))
                        except Exception:  # pragma: no cover
                            # This can't be covered (yet) because I haven't yet found a way to trigger a TarError here.
                            # Also, https://github.com/python/cpython/issues/120740
                            if a.ctx.raise_errors:
                                raise
                            yield UnzipWalkResult(names=new_names, typ=FileType.ERROR)
                    else:
                        yield UnzipWalkResult(names=new_names, typ=FileType.OTHER)
        except Exception:
            if a.ctx.raise_errors:
                raise
            yield UnzipWalkResult(names=a.fns, typ=FileType.ERROR)
        else:
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    elif bl.endswith('.zip'):
        try:
            with ZipFile(a.fh) as zf:
                for zi in zf.infolist():
                    # Note the ZIP specification requires forward slashes for path separators.
                    # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT
                    new_names = (*a.fns, PurePosixPath(zi.filename))
                    if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
                        yield UnzipWalkResult(names=new_names, typ=FileType.SKIP)
                    # Manually detect symlinks in ZIP files (should be rare anyway)
                    # e.g. from zipfile.py: z_info.external_attr = (st.st_mode & 0xFFFF) << 16
                    # we're not going to worry about other special file types in ZIP files
                    elif zi.create_system==3 and stat.S_ISLNK(zi.external_attr>>16):  # 3 is UNIX
                        yield UnzipWalkResult(names=new_names, typ=FileType.SYMLINK)
                    elif zi.is_dir():
                        yield UnzipWalkResult(names=new_names, typ=FileType.DIR)
                    else:  # (note this interface doesn't have an is_file)
                        try:
                            with zf.open(zi) as fh2:  # always binary mode
                                yield from _proc_file(FileProcessorArgs(fns=new_names, fh=fh2, size=zi.file_size, ctx=a.ctx))
                        except Exception:
                            if a.ctx.raise_errors:
                                raise
                            yield UnzipWalkResult(names=new_names, typ=FileType.ERROR)
        except Exception:
            if a.ctx.raise_errors:
                raise
            yield UnzipWalkResult(names=a.fns, typ=FileType.ERROR)
        else:
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    elif bl.endswith('.7z'):
        if W7Z:  # cover-req-lt3.14
            yield from W7Z.process_7z(a, _proc_file)
        else:  # cover-req-ge3.14
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    elif bl.endswith('.bz2'):
        new_names = (*a.fns, a.fns[-1].with_suffix(''))
        if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
            yield UnzipWalkResult(names=a.fns, typ=FileType.SKIP)
        else:
            with BZ2File(a.fh, mode='rb') as fh2:  # always binary, but specify explicitly for clarity
                yield from _proc_file(FileProcessorArgs(fns=new_names, fh=fh2, size=None, ctx=a.ctx))
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    elif bl.endswith('.xz'):
        new_names = (*a.fns, a.fns[-1].with_suffix(''))
        if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
            yield UnzipWalkResult(names=a.fns, typ=FileType.SKIP)
        else:
            with LZMAFile(a.fh, mode='rb') as fh2:  # always binary, but specify explicitly for clarity
                yield from _proc_file(FileProcessorArgs(fns=new_names, fh=fh2, size=None, ctx=a.ctx))
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    elif bl.endswith('.gz'):
        new_names = (*a.fns, a.fns[-1].with_suffix(''))
        if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
            yield UnzipWalkResult(names=a.fns, typ=FileType.SKIP)
        else:
            with GzipFile(fileobj=a.fh, mode='rb') as fh2:  # always binary, but specify explicitly for clarity
                # NOTE casting GzipFile to IO[bytes] isn't 100% safe because the former doesn't
                # implement the full interface, but testing seems to show it's ok...
                yield from _proc_file(FileProcessorArgs(fns=new_names, fh=cast(IO[bytes], fh2), size=None, ctx=a.ctx))
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)
    else:
        assert a.fh.readable(), a.fh  # expected by ReadOnlyBinary
        yield UnzipWalkResult(names=a.fns, typ=FileType.FILE, hnd=a.fh, size=a.size)

def unzipwalk(paths :AnyPaths, *, matcher :Optional[FilterType] = None, raise_errors :bool = True) -> Generator[UnzipWalkResult, None, None]:
    """This generator recursively walks into directories and compressed files and yields named tuples of type :class:`UnzipWalkResult`.

    :param paths: A filename or iterable of filenames.

    :param matcher: When you provide this optional argument, it must be a callable that accepts a sequence of paths
        as its only argument, and returns a boolean value whether this filename should be further processed or not.
        If a file is skipped, a :class:`UnzipWalkResult` of type :class:`FileType.SKIP<FileType>` is yielded.

        *Be aware* that within Zip and tar archives, all files are basically a flat list, so if your matcher
        excludes a directory inside an archive, it must also exclude all files within that directory as well.
        This behavior is different for physical directories in the file system: if you exclude a directory there,
        it will not be descended into, so you won't have to exclude the files inside (though it's good practice
        to write your matcher to exclude them anyway - see for example :meth:`~pathlib.PurePath.is_relative_to`).

    :param raise_errors: When this is turned on (the default), any errors are raised immediately,
        aborting the iteration. If this is turned off, when decompression errors occur, a
        :class:`UnzipWalkResult` of type :class:`FileType.ERROR<FileType>` is yielded for those files instead.

    .. note:: If :mod:`py7zr` is not installed, those archives will not be descended into.

    .. note:: Do not rely on the order of results! But see also the discussion in the main documentation about why
        e.g. ``sorted(unzipwalk(...))`` automatically closes files and so may not be what you want.

    :raises Exception: Because of the various underlying libraries, both this function and :func:`recursive_open` can raise
        a variety of exceptions: :exc:`zipfile.BadZipFile`, :exc:`tarfile.TarError`, ``py7zr.exceptions.ArchiveError``
        and its subclasses like :exc:`py7zr.Bad7zFile`, :exc:`gzip.BadGzipFile`, :exc:`zlib.error`, :exc:`lzma.LZMAError`,
        :exc:`EOFError`, various :exc:`OSError`\\s, and other exceptions may be possible. Therefore, you may need to catch
        all :exc:`Exception`\\s to play it safe.

    .. important:: Errors from :mod:`gzip`, :mod:`bz2`, and :mod:`lzma` (``.gz``, ``.bz2``, and ``.xz`` files,
        respectively) may not be raised until the file is actually read, so you'll probably also want to add an
        exception handler around your ``read()`` call!
    """
    def handle(p :Path):
        try:
            if matcher is not None and not matcher((p,)):
                yield UnzipWalkResult(names=(p,), typ=FileType.SKIP).validate()
            elif p.is_symlink():
                yield UnzipWalkResult(names=(p,), typ=FileType.SYMLINK).validate()  # cover-not-win32
            elif p.is_dir():
                yield UnzipWalkResult(names=(p,), typ=FileType.DIR).validate()
            elif p.is_file():
                with p.open('rb') as fh:
                    yield from ( r.validate() for r in
                        _proc_file(FileProcessorArgs(fns=(p,), fh=fh, size=p.stat().st_size,
                            ctx=ProcessCallContext(matcher=matcher, raise_errors=raise_errors))) )
            else:
                yield UnzipWalkResult(names=(p,), typ=FileType.OTHER).validate()  # cover-not-win32
        except Exception:  # cover-only-linux  # e.g. PermissionError
            if raise_errors:
                raise
            yield UnzipWalkResult(names=(p,), typ=FileType.ERROR).validate()
    for p in to_Paths(paths):
        try:
            is_dir = p.resolve(strict=True).is_dir()
        except Exception:
            if raise_errors:
                raise
            yield UnzipWalkResult(names=(p,), typ=FileType.ERROR).validate()
        else:
            if is_dir:
                for pa in p.rglob('*'):
                    yield from handle(pa)
            else:
                yield from handle(p)



"""
Tests for :mod:`unzipwalk.utils`
================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import unittest
import unzipwalk.utils as uut

class TestUtils(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_decode_tuple(self):
        self.assertEqual( uut.decode_tuple(repr(('hi',))), ('hi',) )
        self.assertEqual( uut.decode_tuple(repr(('hi','there'))), ('hi','there') )
        self.assertEqual( uut.decode_tuple('( "foo" , \'bar\' ) '), ('foo','bar') )
        self.assertEqual( uut.decode_tuple("('hello',)"), ('hello',) )
        self.assertEqual( uut.decode_tuple('"foo","bar"'), ('foo','bar') )
        with self.assertRaises(ValueError):
            uut.decode_tuple('')
        with self.assertRaises(ValueError):
            uut.decode_tuple('X=("foo",)')
        with self.assertRaises(ValueError):
            uut.decode_tuple('(')
        with self.assertRaises(ValueError):
            uut.decode_tuple('()')
        with self.assertRaises(ValueError):
            uut.decode_tuple('("foo")')
        with self.assertRaises(ValueError):
            uut.decode_tuple('("foo","bar",3)')
        with self.assertRaises(ValueError):
            uut.decode_tuple('("foo","bar",str)')
        with self.assertRaises(ValueError):
            uut.decode_tuple('("foo","bar","x"+"y")')
        with self.assertRaises(ValueError):
            uut.decode_tuple('["foo","bar"]')



"""
Definitions for :mod:`unzipwalk`
================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import io
import re
import enum
import hashlib
from contextlib import AbstractContextManager
from collections.abc import Callable, Sequence, Generator
from pathlib import PurePosixPath, PurePath, PureWindowsPath
from typing import Optional, Protocol, NamedTuple, runtime_checkable, IO
from .utils import decode_tuple

class FileType(enum.IntEnum):
    """Used in :class:`UnzipWalkResult` to indicate the type of the file.

    .. warning:: Don't rely on the numeric value of the enum elements, they are automatically generated and may change!
    """
    #: A regular file.
    FILE = enum.auto()
    #: An archive file, will be descended into.
    ARCHIVE = enum.auto()
    #: A directory.
    DIR = enum.auto()
    #: A symbolic link.
    SYMLINK = enum.auto()
    #: Some other file type (e.g. FIFO).
    OTHER = enum.auto()
    #: A file was skipped due to the ``matcher`` filter.
    SKIP = enum.auto()
    #: An error was encountered with this file, when the ``raise_errors`` option is off.
    ERROR = enum.auto()

@runtime_checkable
class ReadOnlyBinary(Protocol):  # pragma: no cover  (b/c Protocol class)
    """Interface for the file handle (file object) used in :class:`UnzipWalkResult`.

    This is essentially the intersection of what the underlying objects support."""
    def close(self) -> None:
        """Close the file.

        .. note::
            :func:`unzipwalk` automatically closes files.
        """
    @property
    def closed(self) -> bool: ...
    def readable(self) -> bool:
        return True
    def read(self, n: int = -1, /) -> bytes: ...
    def readline(self, limit: int = -1, /) -> bytes: ...
    def seekable(self) -> bool: ...
    def seek(self, offset: int, whence: int = io.SEEK_SET, /) -> int: ...

CHECKSUM_LINE_RE = re.compile(r'^([0-9a-f]+) \*(.+)$')
CHECKSUM_COMMENT_RE = re.compile(r'^# ([A-Z]+) (.+)$')

class UnzipWalkResult(NamedTuple):
    """Return type for :func:`unzipwalk`."""
    #: A tuple of the filename(s) as :mod:`pathlib` objects. The first element is always the physical file in the file system.
    #: If the tuple has more than one element, then the yielded file is contained in a compressed file, possibly nested in
    #: other compressed file(s), and the last element of the tuple will contain the file's actual name.
    names :tuple[PurePath, ...]
    #: A :class:`FileType` value representing the type of the current file.
    typ :FileType
    #: When :attr:`typ` is :class:`FileType.FILE<FileType>`, this is a :class:`ReadOnlyBinary` file handle (file object)
    #: for reading the file contents in binary mode. Otherwise, this is :obj:`None`.
    #: If this object was produced by :meth:`from_checksum_line`, this handle will read the checksum of the data, *not the data itself!*
    hnd :Optional[ReadOnlyBinary] = None
    #: When :attr:`typ` is :class:`FileType.FILE<FileType>` or :class:`FileType.ARCHIVE<FileType>`, this field *may* hold the size of the
    #: file, if the compression format and library support knowing the compressed file's size in advance. Otherwise, this is :obj:`None`.
    size :Optional[int] = None

    def validate(self):
        """Validate whether the object's fields are set properly and throw errors if not.

        Intended for internal use, mainly when type checkers are not being used.
        :func:`unzipwalk` validates all the results it returns.

        :return: The object itself, for method chaining.
        :raises ValueError, TypeError: If the object is invalid.
        """
        if not self.names:
            raise ValueError('names is empty')
        if not all( isinstance(n, PurePath) for n in self.names ):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError(f"invalid names {self.names!r}")
        if not isinstance(self.typ, FileType):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError(f"invalid type {self.typ!r}")
        if self.typ==FileType.FILE and not isinstance(self.hnd, ReadOnlyBinary):
            raise TypeError(f"invalid handle {self.hnd!r}")
        if self.typ!=FileType.FILE and self.hnd is not None:
            raise TypeError(f"invalid handle, should be None but is {self.hnd!r}")
        if self.typ not in (FileType.FILE, FileType.ARCHIVE) and self.size is not None:
            raise TypeError(f"invalid size, should be None but is {self.size!r}")
        if self.size is not None and not isinstance(self.size, int):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError(f"invalid size {self.size!r}")
        return self

    def checksum_line(self, hash_algo :str, *, raise_errors :bool = True) -> str:
        """Encodes this object into a line of text suitable for use as a checksum line.

        Intended mostly for internal use by the ``--checksum`` CLI option.
        See :meth:`from_checksum_line` for the inverse operation.

        .. warning:: Requires that the file handle be open (for files), and will read from it to generate the checksum!

        :param hash_algo: The hashing algorithm to use, as recognized by :func:`hashlib.new`.
        :return: The checksum line, without trailing newline.
        """
        names = tuple( str(n) for n in self.names )
        if len(names)==1 and names[0] and names[0].strip()==names[0] and not names[0].startswith('(') \
                and '\n' not in names[0] and '\r' not in names[0]:  # pylint: disable=too-many-boolean-expressions
            name = names[0]
        else:
            name = repr(names)
            assert name.startswith('('), name
        assert '\n' not in name and '\r' not in name, name
        if self.typ == FileType.FILE:
            assert self.hnd is not None, self
            h = hashlib.new(hash_algo)
            try:
                h.update(self.hnd.read())
            except Exception:
                if raise_errors:
                    raise
                return f"# {FileType.ERROR.name} {name}"
            return f"{h.hexdigest().lower()} *{name}"
        return f"# {self.typ.name} {name}"

    @classmethod
    def from_checksum_line(cls, line :str, *, windows :bool=False) -> Optional['UnzipWalkResult']:
        """Decodes a checksum line as produced by :meth:`checksum_line`.

        Intended as a utility function for use when reading files produced by the ``--checksum`` CLI option.

        .. warning:: The ``hnd`` of the returned object will *not* be a handle to
            the data from the file, instead it will be a handle to read the checksum of the file!
            (You could use :func:`recursive_open` to open the files themselves.)

        :param line: The line to parse.
        :param windows: Set this to :obj:`True` if the pathname in the line is in Windows format,
            otherwise it is assumed the filename is in POSIX format.
        :return: The :class:`UnzipWalkResult` object, or :obj:`None` for empty or comment lines.
        :raises ValueError: If the line could not be parsed.
        """
        if not line.strip():
            return None
        path_cls = PureWindowsPath if windows else PurePosixPath
        def mk_names(name :str)-> tuple[PurePath, ...]:
            names = decode_tuple(name) if name.startswith('(') else (name,)
            return tuple(path_cls(p) for p in names)
        if line.lstrip().startswith('#'):  # comment, be lenient to allow user comments
            if m := CHECKSUM_COMMENT_RE.match(line):
                if m.group(1) in FileType.__members__:
                    return cls( names=mk_names(m.group(2)), typ=FileType[m.group(1)] )
            return None
        if m := CHECKSUM_LINE_RE.match(line):
            return cls( names=mk_names(m.group(2)), typ=FileType.FILE, hnd=io.BytesIO(bytes.fromhex(m.group(1))) )
        raise ValueError(f"failed to decode checksum line {line!r}")

# internal types:

FilterType = Callable[[Sequence[PurePath]], bool]

class ProcessCallContext(NamedTuple):
    matcher :Optional[FilterType]
    raise_errors :bool

class FileProcessorArgs(NamedTuple):
    ctx :ProcessCallContext
    fns :tuple[PurePath, ...]
    fh :IO[bytes]
    size :Optional[int]

FileProcessor = Callable[[FileProcessorArgs], Generator[UnzipWalkResult, None, None]]

class RecursiveOpenArgs(NamedTuple):
    fns :tuple[PurePath, ...]
    fh :IO[bytes]

RecursiveOpener = Callable[[RecursiveOpenArgs], AbstractContextManager[IO[bytes]]]



"""
``7z`` Support for :mod:`unzipwalk`
===================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
from typing import Union, Optional, cast, BinaryIO, IO
from collections.abc import Generator
from pathlib import PurePosixPath
from io import BytesIO
# If py7zr isn't available, the following line will raise an exception, causing the imports following it to not be executed.
import py7zr     # pylint: disable=import-error,useless-suppression  # pyright: ignore [reportMissingImports]
import py7zr.io  # pylint: disable=import-error,useless-suppression  # pyright: ignore [reportMissingImports]      # cover-req-lt3.14
from .defs import FileType, UnzipWalkResult, FileProcessor, FileProcessorArgs, RecursiveOpener, RecursiveOpenArgs  # cover-req-lt3.14

# spell-checker: ignore getbuffer nbytes

class Py7zBytesIO(py7zr.io.Py7zIO):  # pyright: ignore [reportUntypedBaseClass]  # cover-req-lt3.14
    def __init__(self, buffer :BytesIO):
        self._buffer = buffer
    def write(self, s :Union[bytes, bytearray]) -> int:
        return self._buffer.write(s)
    def read(self, size :Optional[int] = None) -> bytes:
        return self._buffer.read(size)
    def seek(self, offset :int, whence :int = 0) -> int:
        return self._buffer.seek(offset, whence)
    def flush(self) -> None:
        return self._buffer.flush()
    def size(self) -> int:
        return self._buffer.getbuffer().nbytes

class SingleBytesIOFactory(py7zr.io.WriterFactory):  # pyright: ignore [reportUntypedBaseClass]  # cover-req-lt3.14
    def __init__(self):
        self._filename :Optional[str] = None
        self._buffer :Optional[BytesIO] = None
    def create(self, filename :str) -> py7zr.io.Py7zIO:
        if not isinstance(filename, str):  # pyright: ignore [reportUnnecessaryIsInstance]
            raise TypeError()
        if self._filename is not None or self._buffer is not None:
            raise FileExistsError(f"Attempt to create second file on this factory: {filename!r}")
        self._filename = filename
        self._buffer = BytesIO()
        return Py7zBytesIO(self._buffer)
    def get(self) -> tuple[str, BytesIO]:
        if self._filename is None or self._buffer is None:
            raise FileNotFoundError
        return self._filename, self._buffer

class Wrap7Z:  # cover-req-lt3.14

    @staticmethod
    def _read_one(sz :py7zr.SevenZipFile, fn :str) -> BytesIO:
        """Read one file from a 7z archive as a BytesIO object."""
        fact = SingleBytesIOFactory()
        sz.extract(targets=[str(fn)], factory=fact)
        try:
            return fact.get()[1]
        except FileNotFoundError:  # the getter doesn't know the filename, so replace the exception
            raise FileNotFoundError(f"failed to extract {fn}")  # pylint: disable=raise-missing-from

    @staticmethod
    def recursive_open(a :RecursiveOpenArgs, recurse :RecursiveOpener) -> Generator[IO[bytes], None, None]:
        with py7zr.SevenZipFile(cast(BinaryIO, a.fh)) as sz:
            with recurse(RecursiveOpenArgs(fns=a.fns[1:], fh=Wrap7Z._read_one(sz, str(a.fns[1])))) as inner:
                yield inner

    @staticmethod
    def process_7z(a :FileProcessorArgs, recurse :FileProcessor) -> Generator[UnzipWalkResult, None, None]:
        try:
            # The cast from IO[bytes] to BinaryIO should be ok here I think:
            with py7zr.SevenZipFile(cast(BinaryIO, a.fh)) as sz:
                for f7 in sz.list():
                    new_names = (*a.fns, PurePosixPath(f7.filename))
                    if a.ctx.matcher is not None and not a.ctx.matcher(new_names):
                        yield UnzipWalkResult(names=new_names, typ=FileType.SKIP)
                    elif f7.is_directory:
                        yield UnzipWalkResult(names=new_names, typ=FileType.DIR)
                    else:
                        try:
                            bio = Wrap7Z._read_one(sz, f7.filename)
                        except Exception:  # pylint: disable=[duplicate-code]
                            if a.ctx.raise_errors:
                                raise
                            yield UnzipWalkResult(names=new_names, typ=FileType.ERROR)
                        else:
                            yield from recurse(FileProcessorArgs(fns=new_names, fh=bio, size=f7.uncompressed, ctx=a.ctx))
        except Exception:  # pylint: disable=[duplicate-code]
            if a.ctx.raise_errors:
                raise
            yield UnzipWalkResult(names=a.fns, typ=FileType.ERROR)
        else:  # pylint: disable=[duplicate-code]
            yield UnzipWalkResult(names=a.fns, typ=FileType.ARCHIVE, size=a.size)



"""
Tests for :mod:`unzipwalk`
==========================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import os
import io
import sys
import doctest
import unittest
from hashlib import sha1
from lzma import LZMAError
from gzip import BadGzipFile
from tarfile import TarError
from zipfile import BadZipFile
from tempfile import TemporaryDirectory
from pathlib import Path, PurePosixPath, PureWindowsPath
import unzipwalk as uut
from unzipwalk import FileType
from .defs import P7Z_EX, EXPECT_7Z, BAD_ZIPS, ExpectedResult, TestCaseContext, r2e

def load_tests(_loader, tests, _ignore):
    globs :dict = {}
    def doctest_setup(_t :doctest.DocTest):
        globs['_prev_dir'] = os.getcwd()
        os.chdir( Path(__file__).parent/'doctest_wd' )
    def doctest_teardown(_t :doctest.DocTest):
        os.chdir( globs['_prev_dir'] )
        del globs['_prev_dir']
    tests.addTests(doctest.DocTestSuite(uut, setUp=doctest_setup, tearDown=doctest_teardown, globs=globs))
    return tests

class TestUnzipWalk(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_unzipwalk(self):
        with TestCaseContext() as expect:
            self.assertEqual( expect, sorted( map(r2e, uut.unzipwalk(os.curdir) ) ) )
            # and again, definitely without 7z
            prev = uut.W7Z
            try:  # temporarily pretend 7z is not installed
                uut.W7Z = None
                self.assertEqual( [ x for x in expect if x not in EXPECT_7Z ],
                    sorted( map(r2e, uut.unzipwalk(os.curdir) ) ) )
                with self.assertRaises(ImportError):
                    with uut.recursive_open((Path("more.zip"), PurePosixPath("more/stuff/xyz.7z"), PurePosixPath("even.txt"))):
                        pass  # pragma: no cover
            finally:
                uut.W7Z = prev

    def test_unzipwalk_errs(self):
        with self.assertRaises(FileNotFoundError):
            list(uut.unzipwalk('/this_file_should_not_exist'))

    def test_unzipwalk_matcher(self):
        with TestCaseContext() as expect:
            # filter from the initial path list
            self.assertEqual( sorted(
                    [ r for r in expect if r.fns[0].name != 'more.zip' ]
                    + [ ExpectedResult( (Path("more.zip"),), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: p[0].stem.lower()!='more' ) ) ) )
            # filter from zip file
            self.assertEqual( sorted(
                    [ r for r in expect if r.fns[-1].name != 'six.txt' ]
                    + [ ExpectedResult( (Path("more.zip"), PurePosixPath("more/stuff/six.txt")), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: p[-1].name.lower()!='six.txt' ) ) ) )
            # filter a gz file
            self.assertEqual( sorted(
                    [ r for r in expect if not ( r.fns[0].name=='archive.tar.gz' and len(r.fns)>1 and r.fns[1].name == 'world.txt.gz' ) ]
                    + [ ExpectedResult( (Path("archive.tar.gz"), PurePosixPath("archive/world.txt.gz")), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: len(p)<2 or p[-2].as_posix()!='archive/world.txt.gz' ) ) ) )
            # filter a bz2 file
            self.assertEqual( sorted(
                    [ r for r in expect if not ( r.fns[0].name=='formats.tar.bz2' and len(r.fns)>1 and r.fns[1].name == 'bzip2.txt.bz2' ) ]
                    + [ ExpectedResult( (Path("subdir","formats.tar.bz2"), PurePosixPath("formats/bzip2.txt.bz2")), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: len(p)<2 or p[-2].as_posix()!='formats/bzip2.txt.bz2' ) ) ) )
            # filter an xz file
            self.assertEqual( sorted(
                    [ r for r in expect if not ( r.fns[0].name=='formats.tar.bz2' and len(r.fns)>1 and r.fns[1].name == 'lzma.txt.xz' ) ]
                    + [ ExpectedResult( (Path("subdir","formats.tar.bz2"), PurePosixPath("formats/lzma.txt.xz")), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: len(p)<2 or p[-2].as_posix()!='formats/lzma.txt.xz' ) ) ) )
            # filter from tar file
            self.assertEqual( sorted(
                    [ r for r in expect if not ( len(r.fns)>1 and r.fns[1].stem=='abc' ) ]
                    + [ ExpectedResult( (Path("archive.tar.gz"), PurePosixPath("archive/abc.zip")), None, FileType.SKIP, None ) ]
                ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: p[-1].name != 'abc.zip' ) ) ) )
            if P7Z_EX:  # cover-req-lt3.14
                # filter a file from 7z file
                self.assertEqual( sorted(
                        [ r for r in expect if not ( r.fns[0].name=='opt.7z' and len(r.fns)>1 and r.fns[1].name=='wuv.tgz' ) ]
                        + [ ExpectedResult( (Path("opt.7z"), PurePosixPath("thing/wuv.tgz")), None, FileType.SKIP, None ), ]
                    ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: p[-1].name != 'wuv.tgz' ) ) ) )
                # filter a directory from a 7z file
                self.assertEqual( sorted(
                        [ r for r in expect if not ( r.fns[0].name=='opt.7z' and len(r.fns)>1 ) ]
                        + [ ExpectedResult( (Path("opt.7z"), PurePosixPath("thing")), None, FileType.SKIP, None ),
                            ExpectedResult( (Path("opt.7z"), PurePosixPath("thing/wuv.tgz")), None, FileType.SKIP, None ), ]
                    ), sorted( map(r2e, uut.unzipwalk(os.curdir, matcher=lambda p: not ( len(p)>1 and p[1].parts[0] == 'thing' ) ) ) ) )
            else:  # cover-req-ge3.14
                pass

    def test_recursive_open(self):
        with TestCaseContext() as expect:
            for file in expect:
                if file.typ == FileType.FILE:
                    with uut.recursive_open(file.fns) as fh:
                        self.assertEqual( fh.read(), file.data )
            # text mode
            with uut.recursive_open(("archive.tar.gz", "archive/abc.zip", "abc.txt"), encoding='UTF-8') as fh:
                assert isinstance(fh, io.TextIOWrapper)
                self.assertEqual( fh.readlines(), ["One two three\n", "four five six\n", "seven eight nine\n"] )
            # open an archive
            with uut.recursive_open(('archive.tar.gz', 'archive/abc.zip')) as fh:
                assert isinstance(fh, uut.ReadOnlyBinary)
                self.assertEqual( sha1(fh.read()).hexdigest(), '4d6be7a2e79c3341dd5c4fe669c0ca40a8765031' )
            # basic error
            with self.assertRaises(ValueError):
                with uut.recursive_open(()):
                    pass  # pragma: no cover
            # gzip bad filename
            with self.assertRaises(FileNotFoundError):
                with uut.recursive_open(("archive.tar.gz", "archive/world.txt.gz", "archive/bang.txt")):
                    pass  # pragma: no cover
            # bz2 bad filename
            with self.assertRaises(FileNotFoundError):
                with uut.recursive_open(("subdir/formats.tar.bz2","formats/bzip2.txt.bz2","formats/kaboom.txt")):
                    pass  # pragma: no cover
            # xz bad filename
            with self.assertRaises(FileNotFoundError):
                with uut.recursive_open(("subdir/formats.tar.bz2","formats/lzma.txt.xz","formats/kaboom.txt")):
                    pass  # pragma: no cover
            # TarFile.extractfile: attempt to open a directory
            with self.assertRaises(FileNotFoundError):
                with uut.recursive_open(("archive.tar.gz", "archive/test2/")):
                    pass  # pragma: no cover
            if P7Z_EX:  # cover-req-lt3.14
                # 7z bad filename
                with self.assertRaises(FileNotFoundError):
                    with uut.recursive_open(("opt.7z", "bang")):
                        pass  # pragma: no cover
            else:  # cover-req-ge3.14
                pass

    def test_result_validate(self):
        with self.assertRaises(ValueError):
            uut.UnzipWalkResult((), FileType.OTHER, None, None).validate()
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult(('foo',), FileType.OTHER, None, None).validate()  # type: ignore[arg-type]
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult((Path(),), 'foo', None, None).validate()  # type: ignore[arg-type]
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult((Path(),), FileType.FILE, None, None).validate()
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult((Path(),), FileType.OTHER, io.BytesIO(), None).validate()
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult((Path(),), FileType.FILE, io.BytesIO(), 'x').validate()  # type: ignore[arg-type]
        with self.assertRaises(TypeError):
            uut.UnzipWalkResult((Path(),), FileType.OTHER, None, 42).validate()

    def test_checksum_lines(self):
        res = uut.UnzipWalkResult(names=(PurePosixPath('hello'),), typ=FileType.DIR)
        ln = res.checksum_line("md5")
        self.assertEqual( ln, "# DIR hello" )
        self.assertEqual( uut.UnzipWalkResult.from_checksum_line(ln), res )

        res = uut.UnzipWalkResult(names=(PurePosixPath('hello\nworld'),), typ=FileType.DIR)
        ln = res.checksum_line("md5")
        self.assertEqual( ln, "# DIR ('hello\\nworld',)" )
        self.assertEqual( uut.UnzipWalkResult.from_checksum_line(ln), res )

        res = uut.UnzipWalkResult(names=(PurePosixPath('(hello'),), typ=FileType.DIR)
        ln = res.checksum_line("md5")
        self.assertEqual( ln, "# DIR ('(hello',)" )
        self.assertEqual( uut.UnzipWalkResult.from_checksum_line(ln), res )

        res = uut.UnzipWalkResult(names=(PurePosixPath(' hello '),), typ=FileType.DIR)
        ln = res.checksum_line("md5")
        self.assertEqual( ln, "# DIR (' hello ',)" )
        self.assertEqual( uut.UnzipWalkResult.from_checksum_line(ln), res )

        res2 = uut.UnzipWalkResult.from_checksum_line("# DIR C:\\Foo\\Bar", windows=True)
        assert res2 is not None
        self.assertEqual( res2.names, (PureWindowsPath('C:\\','Foo','Bar'),) )

        res = uut.UnzipWalkResult(names=(PurePosixPath('hello'),PurePosixPath('world')),
            typ=FileType.FILE, hnd=io.BytesIO(b'abcdef'))
        ln = res.checksum_line("md5")
        self.assertEqual( ln, "e80b5017098950fc58aad83c8c14978e *('hello', 'world')" )
        res2 = uut.UnzipWalkResult.from_checksum_line(ln)
        assert res2 is not None
        self.assertEqual( res2.names, (PurePosixPath('hello'),PurePosixPath('world')) )
        self.assertEqual( res2.typ, FileType.FILE )
        assert res2.hnd is not None
        self.assertEqual( res2.hnd.read(), bytes.fromhex('e80b5017098950fc58aad83c8c14978e') )

        self.assertIsNone( uut.UnzipWalkResult.from_checksum_line("# I'm just some comment") )
        self.assertIsNone( uut.UnzipWalkResult.from_checksum_line("# FOO bar") )
        self.assertIsNone( uut.UnzipWalkResult.from_checksum_line("  # and some other comment") )
        self.assertIsNone( uut.UnzipWalkResult.from_checksum_line("  ") )

        with self.assertRaises(ValueError):
            uut.UnzipWalkResult.from_checksum_line("e80b5017098950fc58aad83c8c14978g *kaboom")
        with self.assertRaises(ValueError):
            uut.UnzipWalkResult.from_checksum_line("e80b5017098950fc58aad83c8c14978e *(kaboom")

    def test_errors(self):
        with self.assertRaises(BadZipFile):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.zip'))
        with self.assertRaises(TarError):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.tgz'))
        with self.assertRaises(BadGzipFile):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.tgz.gz'))
        with self.assertRaises(EOFError):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.tgz.bz2'))
        with self.assertRaises(BadZipFile):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.zip.gz'))
        with self.assertRaises(BadZipFile):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.zip.bz2'))
        with self.assertRaises(LZMAError):
            list(uut.unzipwalk(BAD_ZIPS/'not_a.zip.xz'))
        if P7Z_EX:  # cover-req-lt3.14
            with self.assertRaises(type(P7Z_EX)):
                list(uut.unzipwalk(BAD_ZIPS/'not_a.7z'))
            with self.assertRaises(type(P7Z_EX)):
                list(uut.unzipwalk(BAD_ZIPS/'bad.7z'))
            with self.assertRaises(FileExistsError):
                list(uut.unzipwalk(BAD_ZIPS/'double.7z'))
        else:  # cover-req-ge3.14
            pass
        with self.assertRaises(RuntimeError):
            list(uut.unzipwalk(BAD_ZIPS/'features.zip'))
        # the following is commented out due to https://github.com/python/cpython/issues/120740
        #with self.assertRaises(TarError):
        #    list(uut.unzipwalk(pth/'bad.tar.gz'))
        self.assertEqual( sorted(
               (r.names, None if r.hnd is None or r.names[0].name in ('not_a.gz','not_a.bz2','not_a.xz') else r.hnd.read(), r.typ)
               for r in uut.unzipwalk( (BAD_ZIPS, Path('does_not_exist')) , raise_errors=False) ),
            sorted( [
                 ( (Path("does_not_exist"),), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.gz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.gz", BAD_ZIPS/"not_a"), None, FileType.FILE ),  # no error until the file is read (tested below)
                 ( (BAD_ZIPS/"not_a.bz2",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.bz2", BAD_ZIPS/"not_a"), None, FileType.FILE ),  # no error until the file is read (tested below)
                 ( (BAD_ZIPS/"not_a.xz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.xz", BAD_ZIPS/"not_a"), None, FileType.FILE ),  # no error until the file is read (tested below)
                 ( (BAD_ZIPS/"not_a.tar",), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.tar.gz",), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.tgz",), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.zip",), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.tgz.gz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.tgz.gz", BAD_ZIPS/"not_a.tgz"), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.tgz.bz2",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.tgz.bz2", BAD_ZIPS/"not_a.tgz"), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.zip.gz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.zip.gz", BAD_ZIPS/"not_a.zip"), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.zip.bz2",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.zip.bz2", BAD_ZIPS/"not_a.zip"), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"not_a.zip.xz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"not_a.zip.xz", BAD_ZIPS/"not_a.zip"), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"features.zip",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"features.zip", PurePosixPath("spiral.pl")), None, FileType.ERROR ),  # unsupported compression method
                 ( (BAD_ZIPS/"features.zip", PurePosixPath("foo.txt")), b'Top Secret\n', FileType.FILE ),
                 ( (BAD_ZIPS/"features.zip", PurePosixPath("bar.txt")), None, FileType.ERROR ),  # encrypted
                 ( (BAD_ZIPS/"bad.tar.gz",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"bad.tar.gz", PurePosixPath("a")), b'One\n', FileType.FILE ),
                 # the following is commented out due to https://github.com/python/cpython/issues/120740
                 #( (pth/"bad.tar.gz", PurePosixPath("b")), None, FileType.ERROR ),  # bad checksum
                 #( (pth/"bad.tar.gz", PurePosixPath("c")), b'Three\n', FileType.FILE ),
                 ( (BAD_ZIPS/"double.7z",), None, FileType.ARCHIVE ),
                 ( (BAD_ZIPS/"bad.7z",), None, FileType.ARCHIVE ),
            ] + (
                [
                 ( (BAD_ZIPS/"double.7z", PurePosixPath("bar.txt")), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"double.7z", PurePosixPath("bar.txt")), None, FileType.ERROR ),
                 ( (BAD_ZIPS/"bad.7z", PurePosixPath("broken.txt")), None, FileType.ERROR ),  # bad checksum
                 ( (BAD_ZIPS/"not_a.7z",), None, FileType.ERROR ),
                ] if P7Z_EX else [
                 ( (BAD_ZIPS/"not_a.7z",), None, FileType.ARCHIVE ),
                ]) ) )
        with self.assertRaises(BadGzipFile):
            for r in uut.unzipwalk((BAD_ZIPS/'not_a.gz')):  # pragma: no branch
                assert r.hnd is not None
                r.hnd.read()
        with self.assertRaises(OSError):
            for r in uut.unzipwalk((BAD_ZIPS/'not_a.bz2')):  # pragma: no branch
                assert r.hnd is not None
                r.hnd.read()
        with self.assertRaises(LZMAError):
            for r in uut.unzipwalk((BAD_ZIPS/'not_a.xz')):  # pragma: no branch
                assert r.hnd is not None
                r.hnd.read()
        if P7Z_EX:  # cover-req-lt3.14
            with self.assertRaises(FileExistsError):
                with uut.recursive_open((BAD_ZIPS/"double.7z", "bar.txt")):
                    pass  # pragma: no cover
        else:  # cover-req-ge3.14
            pass

    @unittest.skipIf(condition=not sys.platform.startswith('linux'), reason='only on Linux')
    def test_errors_linux(self):  # cover-only-linux
        with TemporaryDirectory() as td:
            f = Path(td)/'foo'
            f.touch()
            f.chmod(0)
            with self.assertRaises(PermissionError):
                list(uut.unzipwalk(td))
            self.assertEqual(
                sorted( map(r2e, uut.unzipwalk(td, raise_errors=False) ) ),
                sorted( [ ExpectedResult( (f,), None, FileType.ERROR, None ), ] ) )

    @unittest.skipIf(condition=P7Z_EX is None, reason='only with 7z support')
    def test_wrap7z(self):  # cover-req-lt3.14
        from unzipwalk.wrap7z import Py7zBytesIO, SingleBytesIOFactory  # pylint: disable=import-outside-toplevel
        pio = Py7zBytesIO(io.BytesIO(b'abc'))
        self.assertEqual(pio.size(), 3)
        self.assertEqual(pio.read(), b'abc')
        pio.flush()
        fact = SingleBytesIOFactory()
        with self.assertRaises(TypeError):
            fact.create(123)  # type: ignore[arg-type]



"""
Tests for :mod:`unzipwalk.__main__`
===================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import os
import io
import sys
import hashlib
import unittest
from pathlib import Path
from gzip import BadGzipFile
from tarfile import TarError
from zipfile import BadZipFile
from unittest.mock import patch
from tempfile import TemporaryDirectory
from contextlib import redirect_stdout, redirect_stderr
import unzipwalk.__main__ as uut
from unzipwalk import FileType
from .defs import P7Z_EX, BAD_ZIPS, TestCaseContext, ExpectedResult

# spell-checker: ignore csha rcmd

class TestUnzipWalkCli(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None  # pylint: disable=invalid-name

    def _run_cli(self, argv :list[str]) -> list[str]:
        sys.argv = [os.path.basename(uut.__file__)] + argv
        with (redirect_stdout(io.StringIO()) as out, redirect_stderr(io.StringIO()) as err,
              patch('argparse.ArgumentParser.exit', side_effect=SystemExit) as mock_exit):
            try:
                uut.main()
            except SystemExit:
                pass
        mock_exit.assert_called_once_with(0)
        self.assertEqual(err.getvalue(), '')
        lines = out.getvalue().splitlines()
        lines.sort()
        return lines

    def test_cli(self):
        expect :list[ExpectedResult]
        with TestCaseContext() as expect:
            exp_basic = sorted( f"FILE {tuple(str(n) for n in e.fns)!r}" for e in expect if e.typ==FileType.FILE )
            self.assertEqual( self._run_cli([]), exp_basic )  # basic
            with TemporaryDirectory() as td:  # --outfile
                tf = Path(td)/'foo'
                self.assertEqual( self._run_cli(['--outfile', str(tf)]), [] )
                with tf.open(encoding='UTF-8') as fh:
                    self.assertEqual( sorted(fh.read().splitlines()), exp_basic )
            self.assertEqual( self._run_cli(['--all-files']), sorted(  # basic + all-files
                f"{e.typ.name} {tuple(str(n) for n in e.fns)!r}" for e in expect ) )
            self.assertEqual( self._run_cli(['--dump']), sorted(  # dump
                f"FILE {tuple(str(n) for n in e.fns)!r} {e.data!r}" for e in expect if e.typ==FileType.FILE ) )
            self.assertEqual( self._run_cli(['-da']), sorted(  # dump + all-files
                f"FILE {tuple(str(n) for n in e.fns)!r} {e.data!r}" if e.typ==FileType.FILE
                else f"{e.typ.name} {tuple(str(n) for n in e.fns)!r}" for e in expect ) )
            self.assertEqual( self._run_cli(['--checksum','sha256']), sorted(  # checksum
                f"{hashlib.sha256(e.data).hexdigest()} *{str(e.fns[0]) if len(e.fns)==1 else repr(tuple(str(n) for n in e.fns))}"
                for e in expect if e.data is not None ) )
            self.assertEqual( self._run_cli(['-a','-csha512']), sorted(  # checksum + all-files
                (f"# {e.typ.name} " if e.data is None else f"{hashlib.sha512(e.data).hexdigest()} *")
                + f"{str(e.fns[0]) if len(e.fns)==1 else repr(tuple(str(n) for n in e.fns))}"
                for e in expect ) )
            self.assertEqual( self._run_cli(['-e','world.*','--exclude=*abc*']), sorted(  # exclude
                f"FILE {tuple(str(n) for n in e.fns)!r}" for e in expect if e.typ==FileType.FILE
                and not ( e.fns[-1].name.startswith('world.') or len(e.fns)>1 and e.fns[1].name=='abc.zip' ) ) )

    def test_cli_errors(self):
        os.chdir(BAD_ZIPS)
        self.assertEqual( self._run_cli(['-d','.','does_not_exist']), sorted( [
            "ERROR ('does_not_exist',)",
            "ERROR ('not_a.gz', 'not_a')",
            "ERROR ('not_a.bz2', 'not_a')",
            "ERROR ('not_a.xz', 'not_a')",
            "ERROR ('not_a.tar',)",
            "ERROR ('not_a.tar.gz',)",
            "ERROR ('not_a.tgz',)",
            "ERROR ('not_a.zip',)",
            "ERROR ('not_a.tgz.gz', 'not_a.tgz')",
            "ERROR ('not_a.tgz.bz2', 'not_a.tgz')",
            "ERROR ('not_a.zip.gz', 'not_a.zip')",
            "ERROR ('not_a.zip.bz2', 'not_a.zip')",
            "ERROR ('not_a.zip.xz', 'not_a.zip')",
            "ERROR ('features.zip', 'spiral.pl')",
            "ERROR ('features.zip', 'bar.txt')",
            "FILE ('features.zip', 'foo.txt') b'Top Secret\\n'",
            "FILE ('bad.tar.gz', 'a') b'One\\n'",
            # the following is commented out due to https://github.com/python/cpython/issues/120740
            #"ERROR ('bad.tar.gz', 'b')",
            #"FILE ('bad.tar.gz', 'c') b'Three\\n'",
        ] + ([
            "ERROR ('bad.7z', 'broken.txt')",
            "ERROR ('double.7z', 'bar.txt')",
            "ERROR ('double.7z', 'bar.txt')",
            "ERROR ('not_a.7z',)",
        ] if P7Z_EX else []) ) )
        self.assertEqual( self._run_cli(['-cmd5','.','does_not_exist']), sorted( [
            "# ERROR does_not_exist",
            "# ERROR ('not_a.gz', 'not_a')",
            "# ERROR ('not_a.bz2', 'not_a')",
            "# ERROR ('not_a.xz', 'not_a')",
            "# ERROR not_a.tar",
            "# ERROR not_a.tar.gz",
            "# ERROR not_a.tgz",
            "# ERROR not_a.zip",
            "# ERROR ('not_a.tgz.gz', 'not_a.tgz')",
            "# ERROR ('not_a.tgz.bz2', 'not_a.tgz')",
            "# ERROR ('not_a.zip.gz', 'not_a.zip')",
            "# ERROR ('not_a.zip.bz2', 'not_a.zip')",
            "# ERROR ('not_a.zip.xz', 'not_a.zip')",
            "# ERROR ('features.zip', 'spiral.pl')",
            "# ERROR ('features.zip', 'bar.txt')",
            "f0294cd41b8a0a0c403911bb212d9edf *('features.zip', 'foo.txt')",
            "b602183573352abf933bc7ca85fd0629 *('bad.tar.gz', 'a')",
            # the following is commented out due to https://github.com/python/cpython/issues/120740
            #"# ERROR ('bad.tar.gz', 'b')",
            #"38a460ffb4cfb15460b4b679ce534181 *('bad.tar.gz', 'c')",
        ] + ([
            "# ERROR not_a.7z",
            "# ERROR ('bad.7z', 'broken.txt')",
            "# ERROR ('double.7z', 'bar.txt')",
            "# ERROR ('double.7z', 'bar.txt')",
        ] if P7Z_EX else []) ) )
        with self.assertRaises(BadGzipFile):
            self._run_cli(['-rd','not_a.gz'])
        with self.assertRaises(BadGzipFile):
            self._run_cli(['-rcmd5','not_a.gz'])
        with self.assertRaises(BadZipFile):
            self._run_cli(['-r','not_a.zip'])
        with self.assertRaises(TarError):
            self._run_cli(['-r','not_a.tgz'])
        with self.assertRaises(RuntimeError):
            self._run_cli(['-r','features.zip'])



#!/usr/bin/env python3
"""
Command-Line Interface for :mod:`unzipwalk`
===========================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import hashlib
import argparse
from fnmatch import fnmatch
from pathlib import PurePath, Path
from collections.abc import Sequence
from igbpyutils.file import open_out
import igbpyutils.error
from . import unzipwalk, FileType

def _arg_parser():
    parser = argparse.ArgumentParser('unzipwalk', description='Recursively walk into directories and archives',
        epilog="* Note --exclude currently only matches against the final name in the sequence, excluding path names, "
        "but this interface may change in future versions. For more control, use the library instead of this command-line tool.\n\n"
        f"** Possible values for ALGO: {', '.join(sorted(hashlib.algorithms_available))}")
    parser.add_argument('-a','--all-files', help="also list dirs, symlinks, etc.", action="store_true")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-d','--dump', help="also dump file contents", action="store_true")
    group.add_argument('-c','--checksum', help="generate a checksum for each file**", choices=hashlib.algorithms_available, metavar="ALGO")
    parser.add_argument('-e', '--exclude', help="filename globs to exclude*", action="append", default=[])
    parser.add_argument('-r', '--raise-errors', help="raise errors instead of reporting them in output", action="store_true")
    parser.add_argument('-o', '--outfile', help="output filename")
    parser.add_argument('paths', metavar='PATH', help='paths to process (default is current directory)', nargs='*')
    return parser

def main(argv=None):
    igbpyutils.error.init_handlers()
    parser = _arg_parser()
    args = parser.parse_args(argv)
    def matcher(paths :Sequence[PurePath]) -> bool:
        return not any( fnmatch(paths[-1].name, pat) for pat in args.exclude )
    report = (FileType.FILE, FileType.ERROR)
    with open_out(args.outfile) as fh:
        for result in unzipwalk( args.paths if args.paths else Path(), matcher=matcher, raise_errors=args.raise_errors ):
            if args.checksum:
                if result.typ in report or args.all_files:
                    print(result.checksum_line(args.checksum, raise_errors=args.raise_errors), file=fh)
            else:
                names = tuple( str(n) for n in result.names )
                if result.typ == FileType.FILE and args.dump:
                    assert result.hnd is not None, result
                    try:
                        data = result.hnd.read()
                    except Exception:
                        if args.raise_errors:
                            raise
                        print(f"{FileType.ERROR.name} {names!r}", file=fh)
                    else:
                        print(f"{result.typ.name} {names!r} {data!r}", file=fh)
                elif result.typ in report or args.all_files:
                    print(f"{result.typ.name} {names!r}", file=fh)
    parser.exit(0)

if __name__=='__main__':  # pragma: no cover
    main()



"""
Internal Utilities for :mod:`unzipwalk`
=======================================

Author, Copyright, and License
------------------------------

Copyright (c) 2022-2025 Hauke Dämpfling (haukex@zero-g.net)
at the Leibniz Institute of Freshwater Ecology and Inland Fisheries (IGB),
Berlin, Germany, https://www.igb-berlin.de/

This library is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see https://www.gnu.org/licenses/
"""
import ast

# spell-checker: ignore elts

def decode_tuple(code :str) -> tuple[str, ...]:
    """Helper function to parse a string as produced by :func:`repr` from a :class:`tuple` of one or more :class:`str`.

    :param code: The Python code to parse.
    :return: The :class:`tuple` that was parsed.
    :raises ValueError: If the code could not be parsed or it was not a tuple.
    """
    try:
        tree = ast.parse(code)
    except SyntaxError as ex:
        raise ValueError() from ex
    if not len(tree.body)==1 or not isinstance(tree.body[0], ast.Expr) or not isinstance(tree.body[0].value, ast.Tuple) \
            or not isinstance(tree.body[0].value.ctx, ast.Load) or len(tree.body[0].value.elts)<1:
        raise ValueError(f"failed to decode tuple {code!r}")
    elements :list[str] = []
    for e in tree.body[0].value.elts:
        if not isinstance(e, ast.Constant) or not isinstance(e.value, str):
            raise ValueError(f"failed to decode tuple {code!r}")
        elements.append(e.value)
    return tuple(elements)
